"""
Constants.

"""

DEFAULT_RESOURCE_CACHE_TTL = 60
PUBLISHED_KEY = "x-request-published"
TTL_KEY = "x-request-ttl"
REQUEST_ID_KEY = "x-request-id"
OPAQUE_TAG_KEY = "x-dynatrace"
URI_KEY = "uri"
RECEIPT_HANDLE_KEY = "receipt_handle"
AWS_SNS = "SNS"
AWS_SQS = "SQS"
